#coding=utf-8

from pynput.keyboard import Listener, Key
# 存在的控制按键
ckeys = ['alt', 'alt_gr', 'alt_l', 'alt_r', 'backspace', 'caps_lock', 'cmd', 'cmd_r', 'ctrl', 'ctrl_l', 'ctrl_r', 'delete', 'down', 'end', 'enter', 'esc', 'f1', 'f10', 'f11', 'f12', 'f13', 'f14', 'f15', 'f16', 'f17', 'f18', 'f19', 'f2', 'f20', 'f21', 'f22', 'f23', 'f24', 'f3', 'f4', 'f5', 'f6', 'f7', 'f8', 'f9', 'home', 'insert', 'left', 'media_next', 'media_play_pause', 'media_previous', 'media_volume_down', 'media_volume_mute', 'media_volume_up', 'menu', 'num_lock', 'page_down', 'page_up', 'pause', 'print_screen', 'right', 'scroll_lock', 'shift', 'shift_r', 'space', 'tab', 'up']
# 转换后的控制按键名称
skeys = ['alt_l', 'alt_r', 'alt_l', 'alt_r', 'backspace', 'caps_lock', 'cmd', 'cmd_r', 'ctrl_l', 'ctrl_l', 'ctrl_r', 'delete', 'down', 'end', 'enter', 'esc', 'f1', 'f10', 'f11', 'f12', 'f13', 'f14', 'f15', 'f16', 'f17', 'f18', 'f19', 'f2', 'f20', 'f21', 'f22', 'f23', 'f24', 'f3', 'f4', 'f5', 'f6', 'f7', 'f8', 'f9', 'home', 'insert', 'left', 'media_next', 'media_play_pause', 'media_previous', 'media_volume_down', 'media_volume_mute', 'media_volume_up', 'menu', 'num_lock', 'page_down', 'page_up', 'pause', 'print_screen', 'right', 'scroll_lock', 'shift_l', 'shift_r', 'space', 'tab', 'up']
ckeys2s = {}
for k,v in zip(ckeys,skeys):
    if hasattr(Key, k):
        ckeys2s[getattr(Key,k)] = v

pass
from buildz import Base
import threading
class Keys(Base):
    def char(self, key):
        if key in ckeys2s:
            return ckeys2s[key]
        if hasattr(key, "char"):
            return key.char
        return None
    def stop(self):
        self.lst.stop()
    def init(self, fc=None):
        '''
            callback: fc(char, press=bool)
        '''
        self.th = None
        self.fc = fc
        self.keys= set()
    def press(self, key):
        #print("press", key)
        c = self.char(key)
        #print("press c:", c)
        if c is not None:
            if c in self.keys:
                return
            self.keys.add(c)
            self.fc(c, True)
    def release(self, key):
        c = self.char(key)
        if c is not None:
            if c in self.keys:
                self.keys.remove(c)
            self.fc(c, False)
    def start(self):
        if self.th:
            return
        self.th = threading.Thread(target=self.run, daemon=True)
        self.th.start()
    def run(self):
        with Listener(on_press=self.press, on_release=self.release) as lst:
            self.lst = lst
            lst.join()

pass


#

import numpy as np
from buildz import Base
class Source(Base):
    def init(self, source, sample_rate = 44100):
        import fluidsynth
        #print(f"sample_rate:", sample_rate)
        self.fs = fluidsynth.Synth()
        self.source = source
        self.sfid=self.fs.sfload(source)
        self.sample_rate = sample_rate
        self.do_select=False
    def get(self, sz=None):
        sz = sz or self.sample_rate
        return self.fs.get_samples(sz)
    def read(self, sec=1.0):
        sz = int(self.sample_rate*sec)
        return self.get(sz)
    def check_select(self):
        if not self.do_select:
            self.select()
    def select(self, channel=0, bank=0, preset=0):
        # 选择钢琴音色(通道0, 音色库0, 音色0)
        self.do_select = True
        self.fs.program_select(channel, self.sfid, bank, preset)
    def press(self, key, power=90, channel=0):
        self.check_select()
        self.fs.noteon(channel, key, power)
    def unpress(self, key, channel=0):
        self.check_select()
        self.fs.noteoff(channel, key)
    def close(self):
        self.fs.delete()
        self.fs = None

pass



#coding=utf-8

__version__="0.1.7"
__author__ = "Zzz, emails: 1174534295@qq.com, 1309458652@qq.com"
__doc__="""
key_musicz.keyz 通过pynput库监听按键
key_musicz.sourcez 通过pyfluidsynth读取sf2文件和生成音频数据
key_musicz.soundz 通过pyaudio播放声音
key_musicz.playz 调用其他模块，监听按键，读取sf2文件，播放音乐
"""



#
from buildz import Base
from . import soundz, sourcez
import time,threading
class Play(Base):
    def init(self, source, sound=None, fps=10, sample_rate = 44100):
        if sound is None:
            sound = soundz.Sound(rate=sample_rate)
        if type(source)==str:
            source = sourcez.Source(source, sample_rate)
        self.source = source
        self.sound = sound
        self.fps = fps
        self.running = True
        self.done_run = False
        self.select = self.source.select
        self.press=self.source.press
        self.unpress = self.source.unpress
        self.th = None
    def start(self):
        if self.th is not None:
            return
        self.th = threading.Thread(target=self.loop, daemon=True)
        self.th.start()
    def stop(self):
        self.running = False
    def close(self):
        self.stop()
        while not self.done_run and self.th is not None:
            time.sleep(0.1)
        self.source.close()
        self.sound.close()
    def loop(self):
        self.running = True
        self.done_run = False
        try:
            while self.running:
                dt = self.source.read(1.0/self.fps)
                self.sound.add(dt)
        finally:
            self.done_run = True

pass


#
from buildz import xf, fz, pyz, dz, Args, Base
import os
from . import playz, keyz
def fetch(keys, as_list = False):
    if type(keys) not in (list,tuple):
        #print(f"return not list", keys, as_list)
        if as_list:
            return [keys]
        return keys
    if as_list:
        rst = []
        for it in keys:
            rst+=fetch(it, as_list)
        #print("return list:", rst)
        return rst
    if len(keys)==2:
        if type(keys[0]) not in (list, tuple, dict) and keys[0]!='vals':
            rst = {}
            rst[keys[0]] = keys[1]
            return rst
        elif keys[0]=='vals' or (keys[0][0]=='vals' and len(keys[0])==1):
            rst = []
            for it in keys[1]:
                rst+= fetch(it, as_list)
            return rst
        elif keys[0][0] == 'vals':
            rst = [fetch(it, True) for it in keys[1]]
            ks = keys[0][1:]
            out = []
            for it in rst:
                tmp = {}
                for k,v in zip(ks, it):
                    tmp[k] = v
                out.append(tmp)
            return out
    temp = {}
    arr = []
    for it in keys:
        tmp = fetch(it, as_list)
        if type(tmp) == dict:
            temp.update(tmp)
        else:
            arr += tmp
    if len(arr)==0:
        return temp
    out = []
    for it in arr:
        dt = dict(temp)
        dt.update(it)
        out.append(dt)
    return out

def init(fps, sys_conf={}):
    confs = [xf.loadxf(fp, as_args=True,spc=False) for fp in fps]
    if len(confs)>0:
        conf = confs[0]
    else:
        conf = {}
    for it in confs[1:]:
        dz.deep_fill_argx(it, conf, 1)
    dz.deep_fill_argx(sys_conf, conf, 1)
    if isinstance(conf, Args):
        conf = conf.dicts
    ks, vs, inits= dz.g(conf, keys=[], vars={}, init={})
    if isinstance(ks, Args):
        ks = ks.as_list(cmb=0, item_args = False, deep=True)
    if isinstance(vs, Args):
        vs = vs.as_dict(True)
    if isinstance(inits, Args):
        inits = inits.as_dict(True)
    cmds = []
    for it in ks:
        cs = fetch(it)
        if type(cs) == dict:
            cs = [cs]
        cmds+=cs
    # print("orders:", cmds)
    # print("vars:", vs)
    # print("init:", inits)
    return cmds, vs, inits

pass
class Orders(Base):
    def init(self):
        self.orders = {}
    def set(self, key, fc):
        self.orders[key] = fc
    def call(self, maps, *args):
        maps = dict(maps)
        key = maps['action']
        del maps['action']
        return self.orders[key](*args, **maps)

pass
from .base import default_src, conf_fp, path
class Conf(Base):
    def win_fix(self, libpath):
        #print(f"[TESTZ] libpath: {libpath}")
        if libpath is None:
            return
        import os
        if hasattr(os, 'add_dll_directory'):  # Python 3.8+ on Windows only
            os.add_dll_directory(libpath)
            os.environ['PATH'] += f';{libpath}'
    def init(self, fps, sys_conf={}):
        self.to_stops = set()
        if type(fps) not in (list, tuple):
            fps = [fps]
        #self.play = play
        #print(f"[TESTZ] sys_conf: {sys_conf}")
        self.fps = fps
        cmds, vs, inits = init(fps, sys_conf)   
        #print(f"[TESTZ] inits: {inits}")
        sfile, fps, select, sample_rate, libpath = dz.g(inits, sfile = default_src, fps=30, select={}, sample_rate=44100, libpath=None)
        sfile = path(sfile)
        libpath = path(libpath)
        self.win_fix(libpath)
        sfile = sfile or default_src
        if sfile is None or not os.path.isfile(sfile):
            print(f"ERROR: sf2文件'{sfile}'不存在 / sf2 file '{sfile}' not exists")
            exit()
        play = playz.Play(sfile,fps=fps, sample_rate=sample_rate)
        play.select(**select)
        channel = dz.g(select, channel= 0)
        self.channel= channel
        self.play = play
        self.ks = keyz.Keys(self.press_callback)
        self.vars = vs
        self.baks = {}
        rst = {}
        for cmd in cmds:
            key = str(cmd['key'])
            rst[key] = cmd
        self.keys = rst
        #print(f"keys:", list(self.keys.keys()))
        self.build_fc()
        self.running = True
    def start(self):
        self.running = True
        self.play.start()
        self.ks.start()
    def close(self):
        self.ks.stop()
        self.play.stop()
        self.play.close()
    def stop(self,*a,**b):
        for n in self.to_stops:
            self.play.unpress(n, self.channel)
        self.to_stops = set()
    def quit(self, *a, **b):
        self.running = False
        input("press enter to quit:\n")
    def wait(self):
        import time
        while self.running:
            time.sleep(0.1)
    def press(self, do_press, label, var, val, **maps):
        if do_press:
            bak = self.vars[label][var]
            dz.dset(self.baks, [label, var], bak)
            self.vars[label][var] = val
        else:
            bak = self.baks[label][var]
            self.vars[label][var] = bak
    def change(self, do_press, label, var, val, **_):
        if not do_press:
            return
        self.vars[label][var] = val
    def mode(self, do_press, mode, **_):
        self.vars['mode'] = mode
    def fix_power(self, n,power):
        vmin = dz.dget(self.vars, 'soundfix.min'.split("."), 0)[0]
        vmax = dz.dget(self.vars, 'soundfix.max'.split("."), 0)[0]
        vdst = vmax-vmin
        n = min(max(n, 36), 132)
        rate = (n-36)/(132-36)
        #rate=rate*rate
        return int(power+vmin+vdst*rate)
    def sound(self, do_preess, label, sound, power=None, **_):
        vs = self.vars[label]
        n = vs['base']+sound
        if not do_preess:
            if self.vars['mode']==0:
                self.play.unpress(n, self.channel)
                if n in self.to_stops:
                    self.to_stops.remove(n)
            return
        if power is None:
            power = vs['power']
        if n in self.to_stops:
            self.play.unpress(n, self.channel)
            self.to_stops.remove(n)
        power = self.fix_power(n, power)
        print(f"play.press({n}, {power})")
        self.play.press(n, power, self.channel)
        self.to_stops.add(n)
    def press_callback(self, char, press):
        #print(f"press:", char, press)
        if char not in self.keys:
            return
        maps = self.keys[char]
        self.orders(maps, press)
    def build_fc(self):
        self.orders = Orders()
        self.orders.set('press', self.press)
        self.orders.set('stop', self.stop)
        self.orders.set('mode', self.mode)
        self.orders.set('quit', self.quit)
        self.orders.set('change', self.change)
        self.orders.set('sound', self.sound)


from buildz import argx
s_help = fz.fread(path("help.txt")).decode("utf-8")
def test():
    import time,sys
    ft = argx.Fetch(*xf.loads("[fp,sfile,libpath,default,help],{f:fp,s:sfile,l:libpath,t:default,h:help}"))
    rst = ft(sys.argv[1:])
    if 'help' in rst:
        print(s_help)
        return
    fps = []
    if 'fp' in rst:
        fps.append(path(rst['fp']))
        del rst['fp']
    default = '1'
    if 'default' in rst:
        default = rst['default']
        del rst['default']
    if default in (0,'0'):
        default = None
    elif default in (1,'1'):
        default = conf_fp
    if default:
        fps = [path(default)]+fps
    sys_conf = {'init':rst}
    conf = Conf(fps, sys_conf)
    conf.start()
    print("run success, enter '~' to quit")
    print("运行中,按下'~'键来退出")
    conf.wait()
    print("release")
    conf.close()

pass

pyz.lc(locals(), test)
'''
python -m key_musicz.run -dD:\rootz\python\gits\key_musicz\lib
'''


#coding=utf-8

import pyaudio
import numpy as np
import math

nbyte = 2
ndtype = np.int16

p = pyaudio.PyAudio()
def create(rate):
    return p.open(format=p.get_format_from_width(nbyte), channels=1, rate=rate, output=True)

pass

def close(stream):
    stream.stop_stream()
    stream.close()

pass

def release():
    p.terminate()

pass


class Sound:
    def __init__(self, stream = None, rate=327680):
        #print(f"sound rate:", rate)
        if stream is None:
            stream = create(rate)
        self.stream = stream
    def start(self):
        pass
    def add(self, data):
        if type(data)!=bytes:
            data = data.tobytes()
        self.stream.write(data)
    def close(self):
        close(self.stream)
        release()

pass